#define PIC16F676

#define LCD

#include "16f676.h"
#pragma config |= 0x318C // INT OSC; 318C with Watch Dog;

#pragma update_RP 1

#include "types.h"

//--- PORT DEFINITIONS ---------------------------------------------------
//               76543210
#define  TrisA 0b00001111                
//               ||||||||
//               |||||||ACCELX         
//               ||||||ACCELY        
//               |||||SERVO_IN
//               ||||BUTTON
//               |||SERVO_OUT
//               ||LED
//               |---
//               ---

//               76543210
#define  TrisC 0b00000000
//               ||||||||
//               |||||||D4
//               ||||||D5
//               |||||D6 
//               ||||D7
//               |||E 
//               ||A0
//               |---
//               ---

#pragma  bit ACCELX    @ PORTA.0   //        - PIC-13
#pragma  bit ACCELY    @ PORTA.1   //        - PIC-12
#pragma  bit SERVO_IN  @ PORTA.2   //        - PIC-11 (pin of INTerrupt)
#pragma  bit BUTTON    @ PORTA.3   //        - PIC-4 
#pragma  bit SERVO_OUT @ PORTA.4   //        - PIC-3 
#pragma  bit LED       @ PORTA.5   //        - PIC-2 

#pragma  bit LCD_A0   @ PORTC.5   // LCD-4  - PIC-5 
#pragma  bit LCD_E    @ PORTC.4   // LCD-6  - PIC-6 
#pragma  bit LCD_D4   @ PORTC.0   // LCD-11 - PIC-10
#pragma  bit LCD_D5   @ PORTC.1   // LCD-12 - PIC-9
#pragma  bit LCD_D6   @ PORTC.2   // LCD-13 - PIC-8
#pragma  bit LCD_D7   @ PORTC.3   // LCD-14 - PIC-7

#define BUTTON_PRESSED 0

//------------------------------------------------------------------------

#include "int16CXX.H"

// Servo[0] = Horizontal pan; Servo[1]= Vertical tilt; Servo[2]=AP shutter sw.;
#define NUM_SERVO 3
LONG Servo[NUM_SERVO];

#define NUM_SERVO_DATA 4
LONG ServoData_X[NUM_SERVO_DATA-1];
LONG ServoData_Y[NUM_SERVO_DATA-1];
byte CurrentServo;
bit  ServoStart;
bit  WaitForGap;
LONG FramesCount;
LONG BrokenFramesCount;

#define SERVO_PERIOD	20000
#define SERVO_NEUTRAL	1400 // 1500 - strange but have to it to eliminate trimming;
#define SERVO_SPAN	500
#define SERVO_MAX (SERVO_NEUTRAL+SERVO_SPAN)
#define SERVO_MIN (SERVO_NEUTRAL-SERVO_SPAN)
#define SERVO_START	300
#define SERVO_GAP	(3*SERVO_MAX)

// extended throws for my BMS-300 9g servo;
#define SERVO_EXT_SPAN  750
#define SERVO_EXT_MAX (SERVO_NEUTRAL+SERVO_EXT_SPAN)
#define SERVO_EXT_MIN (SERVO_NEUTRAL-SERVO_EXT_SPAN)

#define START_CH	4 // Servo[0] = Rudder(Ch.4); Servo[1]= Switch(Ch.5); Servo[2]=(Ch.6);

#define SERVO_HIGH	0 // 0=inverse servo polarity (Hitec Flash 5 needs 0);

void StartWaitForGap();

#pragma origin 4
interrupt serverX( void)
{
/* W and STATUS are saved by the next macro. PCLATH is also
   saved if necessary. The code produced is CPU-dependent. */

int_save_registers    // W, STATUS (and PCLATH if required)
char sv_FSR = FSR;  // save FSR if required
char sv_PCLATH = PCLATH; // save PCLATH

  // handle the interrupt (insert application code here)

// INTE enabled;
// INTEDG = 1 (0->1);
// needs 3-4 cycles to get into interrupt;

   if (INTF == 1) { // --------- SERVO_EDGE_Interrupt 

     if (WaitForGap) { // found impulse during waiting gap -> need to restart WaitForGap;

       BrokenFramesCount++;
       StartWaitForGap();

     } else {

       if (CurrentServo < START_CH) { // 0, 1, 2, 3
         if (CurrentServo == 0) 
            FramesCount++;

         // start new synchro;
         SERVO_OUT = SERVO_HIGH;

         T1CON = 0b00000000; // Stop timer;
         TMR1H = ((65536-SERVO_START) >> 8) & 0xFF; 
         TMR1L = (65536-SERVO_START) & 0xFF; // start counting;

         T1CON = 0b00000001; // Resume counting;
         CurrentServo++; 
       }

     } 

     INTF = 0;  /* reset flag */

   } else { // ----------- TMR1_Interrupt 
   
     if (TMR1IF) {

       T1CON = 0b00000000; // Stop timer;

       if (WaitForGap) { // Gap found, because timer finishes count and no INT occurs;
         // Timer1 is inactive - we wait for edge INT;
         WaitForGap = 0;  // now we sit inside gap;
         CurrentServo = 0; // start of sequence;

       } else {

         SERVO_OUT = !SERVO_HIGH;

         if (CurrentServo >= START_CH) { // 4

          if (CurrentServo == NUM_SERVO+START_CH) { // 2+4=6
            // we finished short "1" impulse follewed by last ch., so we will wait the gap;
            StartWaitForGap();

          } else {

            LONG width;
            if (ServoStart == 1) {

              width = SERVO_START;
              SERVO_OUT = SERVO_HIGH;
              ServoStart = 0;
              CurrentServo++;

            } else {
              width = Servo[CurrentServo-START_CH]; // 4.. -> 0..
              width -= SERVO_START;

              ServoStart = 1; // followed by short "1" for channel delimiter; 

              if (width == 0)
                width.low8 = 1;
            }

            width = -width;
            TMR1H = width.high8;
            TMR1L = width.low8;

            T1CON = 0b00000001; // Resume counting;
          }

         }

       }

       TMR1IF = 0;         
     }

   }

PCLATH = sv_PCLATH;    // restore PCLATH
FSR = sv_FSR;       // restore FSR if saved
int_restore_registers // W, STATUS (and PCLATH if required)
}

void StartWaitForGap() {
  WaitForGap = 1; 
  // find 3 slot silence at least in TX sending;
  T1CON = 0b00000000; // Stop timer;
  TMR1H = ((65536-SERVO_GAP) >> 8) & 0xFF; 
  TMR1L = (65536-SERVO_GAP) & 0xFF; // start counting;
  T1CON = 0b00000001; // Resume counting;
}

//#include "BIN\MATH16.H"

#include "delay.c"

#include "lcd.c"

void InitServo() {
  // init channels;
  byte i;
  for (i=0; i<NUM_SERVO; i++)
    Servo[i]= SERVO_NEUTRAL;
  for (i=0; i<NUM_SERVO_DATA-1; i++) {
    ServoData_X[i]= SERVO_NEUTRAL;
    ServoData_Y[i]= SERVO_NEUTRAL;
  }
}

//--- ANALOG SETUP DATA --------------------------------------------------
#define ADChannel_ACCELX 	0
#define ADChannel_ACCELY	1

LONG Accel[2];
bit  button;

#define NUM_VOLTAGE_SAMPLES     8  // 2, 4, 8, 16

LONG ReadADC(byte piADChannel) {
  byte a = (piADChannel << 2);
  ADCON0 = (ADCON0 & 0b111.000.11) | a;
  Delay100us(10);
  GO = 1;
  while (GO) { };

  long unsigned int u;
  u.high8 = ADRESH;
  u.low8 = ADRESL;
  return u;
}

void ReadChannels() {
  Accel[0] = 0;
  Accel[1] = 0;
  signed int i = NUM_VOLTAGE_SAMPLES;
  while (i > 0) {
    Accel[0] += ReadADC(ADChannel_ACCELX);
    Accel[1] += ReadADC(ADChannel_ACCELY);
    i--;
  }

  Accel[0] /= NUM_VOLTAGE_SAMPLES; // 1:1
  Accel[1] /= NUM_VOLTAGE_SAMPLES; // 1:1
  button = BUTTON == BUTTON_PRESSED;
}

#define ACCEL_STABLE 	536
#define MAX_ACCEL_X	100
#define MAX_ACCEL_Y	50
#define SAFETY_MARGIN	15

// Find if channels Cell1, Cell2, Cell3 and Temp are available;
void ReadAndOutputAvailableChannels() {
  ReadChannels();

#ifdef LCD
  LCD_CURSOR_LINE_1 (0);  WriteNum (Accel[0]); // Vertical movement;
  LCD_CURSOR_LINE_1 (4);  WriteNum (Accel[1]); // Horizontal movement;
#endif

  int ScaledAccel[2];

  // Vertical movement;
  if (Accel[0] > (ACCEL_STABLE-SAFETY_MARGIN))
    ScaledAccel[0] = 0;
  else {
    Accel[0] = (ACCEL_STABLE-SAFETY_MARGIN) - Accel[0];
    if (Accel[0] > MAX_ACCEL_X)
      ScaledAccel[0] = MAX_ACCEL_X;
    else
      ScaledAccel[0] = (int) Accel[0];
  }

  // Horizontal movement;
  if (Accel[1] < (ACCEL_STABLE-SAFETY_MARGIN)) {
    Accel[1] = (ACCEL_STABLE-SAFETY_MARGIN) - Accel[1];
    if (Accel[1] > MAX_ACCEL_Y)
      ScaledAccel[1] = -MAX_ACCEL_Y;
    else
      ScaledAccel[1] = -((int) Accel[1]);    
  } else
    if (Accel[1] > (ACCEL_STABLE+SAFETY_MARGIN)) {
      Accel[1] -= (ACCEL_STABLE+SAFETY_MARGIN);
      if (Accel[1] > MAX_ACCEL_Y)
        ScaledAccel[1] = MAX_ACCEL_Y;
      else
        ScaledAccel[1] = (int) Accel[1];    
    } else
      ScaledAccel[1] = 0;
  
  // Vertical movement;
  int x = ScaledAccel[0]-(MAX_ACCEL_X/2); // 0..MAX_ACCEL_X -> -MAX_ACCEL_X/2..MAX_ACCEL_X/2; (-50..+50);
//    signed long x1 = x*SERVO_SPAN/(MAX_ACCEL_X/2); // x*500/50=x*10;
  signed long a = (signed long) x;
  signed long x1 = a*10; // x*500/50=x*10;

  // Horizonal movement;
//    signed long y1 = ScaledAccel[1]*SERVO_SPAN/MAX_ACCEL_Y; // y*500/50 = y*10;
  a = (signed long)ScaledAccel[1];
  signed long y1 = (a)*13;

  // chained feedback filter;
  LONG x2 = (LONG) (x1+SERVO_NEUTRAL), y2=(LONG) (y1+SERVO_NEUTRAL);

  LCD_CURSOR_LINE_2 (8);  WriteNum (x2);
  LCD_CURSOR_LINE_2 (12);  WriteNum (y2);  

  // sum data from NUM_SERVO_DATA-1 previous frames and add x1, y1;
  // shift ServoData arrays towards the begining;
  byte i;

  // add first items before, because they will be shifted out;
  y2 += ServoData_Y[0];  
  x2 += ServoData_X[0];  

  for (i=0; i < NUM_SERVO_DATA-2; i++) { // 0,1,2
    LONG t = ServoData_Y[i+1];
    ServoData_Y[i] = t;
    y2 += t;
    t = ServoData_X[i+1];
    ServoData_X[i] = t;
    x2 += t;
  }
  x2 /= NUM_SERVO_DATA; 
  y2 /= NUM_SERVO_DATA;

  // round by 16 to eliminate tiny movements;
//  x2.low8 &= 0xFFF0;
//  y2.low8 &= 0xFFF0;

  ServoData_Y[NUM_SERVO_DATA-2] = y2;
  ServoData_X[NUM_SERVO_DATA-2] = x2;

  GIE = 0;
  Servo[1] = x2; 
  Servo[0] = y2;
  if (button)
    Servo[2] = SERVO_MIN;
  else 
    Servo[2] = SERVO_MAX;
  GIE = 1;

#ifdef LCD
  LCD_CURSOR_LINE_2 (0);  WriteNum (Servo[0]);
  LCD_CURSOR_LINE_2 (4);  WriteNum (Servo[1]);
  LCD_CURSOR_LINE_2 (8);  WriteNum (BrokenFramesCount);
  LCD_CURSOR_LINE_2 (12);  WriteNum (FramesCount);  
#endif
  LED = button;
}

char GetOsccal();							// Defined at end of file

void main() {
  W = GetOsccal();						// Set built-in oscillator calibration
  OSCCAL = W;
  
  TRISA = TrisA;
  TRISC = TrisC;
  OPTION = 0b11001110; // no pull-up, int/RA2 on 0->1; Prescalder 64 -> ~1 sec. to WDT;
  PORTA = 0; 
  PORTC = 0;
  button = 0;

  CMCON = 0b.0000.0111; // comparator is turned off;

  ANSEL = (1 << ADChannel_ACCELX) | (1 << ADChannel_ACCELY);

  ADCON1 = 0b0.101.0000;      //  Select the Clock as Fosc/16

  ADIE = 0;

  ADCON0 = 0b10.0.000.01;     //  Turn on the ADC
                              //   Bit 7 - Right Justified Sample
                              //   Bit 6 - Dont Use VRef
                              //   Bit 4:2 - Channel 0
                              //   Bit 1 - Do not Start
                              //   Bit 0 - Turn on ADC
  LED=1;

#ifdef LCD
  LCD_Init();          
#endif

  InitServo();
  CurrentServo = 0;
  ServoStart = 0;
  SERVO_OUT = !SERVO_HIGH;
  BrokenFramesCount = 0;
  FramesCount = 0;

  // Init Timer1;
  TMR1IF = 0;           
  INTF = 0;

  StartWaitForGap();

  PIE1 = 1; 

  PEIE = 1;
  INTE = 1;
  GIE = 1;

  ReadAndOutputAvailableChannels();

  Delay10ms(20);
  LED=0;
  Delay10ms(20);
  LED=1;
  Delay10ms(20);
  LED=0;

  while (1) {
    clrwdt(); 
    ReadAndOutputAvailableChannels();

  };
}

#pragma origin 0x3FF
// This is a dummy - actual procedure is preset in chip and will override
// A change of 1 causes about 1% change in frequency
char GetOsccal() {return 0x0;}

/*
// Initialize EEPROM data
#define EEPROM_START 0x2100
#pragma cdata[EEPROM_START]
#pragma cdata[] = MAGIC, 0x00	
*/
