#ifdef LCD

// LCD connector;
// 1   2   3   4   5   6   7   8   9   
// GND U+  A0  E   D4  D5  D6  D7  UBL+

//--- LCD FUNCTIONS ------------------------------------------------------

#define LCD_A0_mask 0b00100000
#define LCD_DD_mask 0b00001111

#define LCD_DELAY_BASE_100_US 1

void lcdWrite (char c) {
  // high data byte;
  c = swap(c);
  PORTC = (c & LCD_DD_mask) | LCD_A0_mask; LCD_E = 1; nop(); LCD_E = 0;
  Delay100us (LCD_DELAY_BASE_100_US);
  // low data byte;
  c = swap(c);
  PORTC = (c & LCD_DD_mask) | LCD_A0_mask; LCD_E = 1; nop(); LCD_E = 0;
  Delay100us (LCD_DELAY_BASE_100_US);
}

void lcdCommand (char c) {
  // high data byte;
  c = swap(c);
  PORTC = (c & LCD_DD_mask); LCD_E = 1; nop(); LCD_E = 0;
  Delay100us (LCD_DELAY_BASE_100_US);
  // low data byte;
  c = swap(c);
  PORTC = (c & LCD_DD_mask); LCD_E = 1; nop(); LCD_E = 0;
  Delay100us (LCD_DELAY_BASE_100_US);
}

void lcdCommand8 (char c) {
  PORTC = c; LCD_E = 1; nop(); LCD_E = 0;
  Delay100us (LCD_DELAY_BASE_100_US);
}

#define LCD_CMD_POS1		( 0x80 )
#define LCD_CMD_POS2		(0x80 | 0x40)

#define LCD_CURSOR_ON		lcdCommand (0xC | 0x2);
#define LCD_CURSOR_OFF		lcdCommand (0xC);
#define LCD_CURSOR_BLINK_ON	lcdCommand (0xC | 0x2 | 0x1);
#define LCD_CLEAR		{ lcdCommand (0x1); Delay100us (15+LCD_DELAY_BASE_100_US); }
#define LCD_CURSOR_HOME		lcdCommand (0x2);
#define LCD_SHIFT_CURSOR_LEFT	lcdCommand (0x10);
#define LCD_SHIFT_CURSOR_RIGHT	lcdCommand (0x10 | 0x4);
#define LCD_SHIFT_DISPLAY_LEFT	lcdCommand (0x10 | 0x8);
#define LCD_SHIFT_DISPLAY_RIGHT	lcdCommand (0x10 | 0x8 | 0x4);
#define LCD_CURSOR_LINE_1(x)	lcdCommand (LCD_CMD_POS1 | x);
#define LCD_CURSOR_LINE_2(x)	lcdCommand (LCD_CMD_POS2 | x);

void lcdProgram (const char * p) {
  byte i;
  BOOL lbCommand = 1;
  char c;
  for (i = 0; ; i++) {
    char c = p[i];
    if (c == 0)
      return;
    if (lbCommand) {
      lcdCommand (c);
      if (c & LCD_CMD_POS1) // output character after cursor placement command;
        lbCommand = 0;
    } else {
      lcdWrite (c);
      lbCommand = 1;
    }
  }
}

void lcdWriteln (const char * p) {
  byte i;
  for (i = 0; p[i] != 0; i++) {
      lcdWrite(p[i]);
  }
}

const char LCD_Data_Init[] = {
  0x28, //  
  0x0C, //  
  0x6,  //     
  0x0F, //     
//  0x80, //  
//  ' ',
  0x01,  // ,  
  0
};

void LCD_Init() { // 16s2d
  PORTC = 0;
  Delay10ms (5);

// 30h  
  lcdCommand8 (3);
  lcdCommand8 (3);
  lcdCommand8 (3);

// 20h  
  lcdCommand8 (2);

//4bit
  lcdProgram (LCD_Data_Init);
  Delay100us (15);
}

void WriteNum (LONG n) {
  byte i = '0';
  while (n >= 1000) {
    n -= 1000;
    i++;
  }
  lcdWrite (i);

  i = '0';
  while (n >= 100) {
    n -= 100;
    i++;
  }
  lcdWrite (i);

  i = '0';
  while (n >= 10) {
    n -= 10;
    i++;
  }
  lcdWrite (i);
  lcdWrite (n+'0');
}

void WriteSNum (int n) {
  if (n < 0) {
    lcdWrite ('-');
    n = -n;    
  } else {
    lcdWrite ('+');
  }

  byte i = '0';
  while (n >= 100) {
    n -= 100;
    i++;
  }
  lcdWrite (i);

  i = '0';
  while (n >= 10) {
    n -= 10;
    i++;
  }
  lcdWrite (i);
  lcdWrite (n+'0');
}

#endif

